import sys
import os
import re
import common
forge = sys.modules["forge"]

def setup(module):
	manifest = 'ext/alembic/manifest.cc'
	with open(manifest, 'w') as outfile:
		outfile.write("")

	for alembic_source in forge.alembic_sources:
		if not os.path.exists(alembic_source[2]) or not os.path.exists(alembic_source[3]):
			sys.stdout.write("-" + alembic_source[0] + " ")
			continue

		alembic_lib = common.find_alembic_lib(alembic_source[3])
		if alembic_lib == '':
			sys.stdout.write("-" + alembic_source[0] + " ")
			continue

		if alembic_source[0] == "":
			sys.stdout.write("/ ")
		else:
			sys.stdout.write(alembic_source[0] + " ")

#		sys.stdout.write(alembic_lib + " ")

		srcList = [	"SurfaceAccessibleAbc",
					"alembic.pmh",
					"alembicDL" ]

		variant = alembic_source[0]
		if variant != "":
			variantPath = module.modPath + '/' + variant
			if os.path.lexists(variantPath) == 0:
				os.symlink('.', variantPath)

			srcListVariant = []
			for src in srcList:
				srcListVariant += [ variant + '/' + src ]
			srcList = srcListVariant

		dllname = "fexAlembicDL" + variant

		with open(manifest, 'a') as outfile:
			suffix = ""
			if variant != "":
				suffix = "." + variant
			outfile.write('\tspManifest->catalog<String>(\n'+
					'\t\t\t"SurfaceAccessibleI.SurfaceAccessibleAbc.fe.abc'+
					suffix+'")=\n'+
					'\t\t\t"fexAlembic'+variant+'";\n');

		dll = module.DLL( dllname, srcList )

		for src in srcList:
			relativePath = forge.RelativePath(forge.rootPath,module.modPath)
			modObjPath = os.path.join(forge.objPath, relativePath)
			objName = modObjPath + '/' + src
			if re.compile(r'(.*)\.pmh$').match(src):
				objName += '.gch'
			else:
				objName += forge.objsuf
			srcTarget = forge.FindTarget(objName)
			srcTarget.includemap = { 'alembic' : alembic_source[2] }

			alembic_openexr = alembic_source[2] + '/OpenEXR'
			if os.path.exists(alembic_openexr):
				srcTarget.includemap['openexr'] = alembic_source[2] + '/OpenEXR'
			elif forge.openexr_source != '':
				default_openexr = forge.openexr_source + '/include/OpenEXR'
				if os.path.exists(default_openexr):
					srcTarget.includemap['openexr'] = default_openexr

		dll.linkmap = { "alembic" :
				"-L" + alembic_source[3] + ' -l' + alembic_lib }
		if alembic_source[0] == '':
			dll.linkmap["alembic"] += ' -Xlinker -rpath ' + alembic_source[3]

#		dll.linkmap['alembic'] += ' -lHalf'
#		dll.linkmap['alembic'] += ' -lhdf5'
		dll.linkmap['alembic'] += ' -lAlembicAbc'
#		dll.linkmap['alembic'] += ' -lAlembicAbcCollection'
		dll.linkmap['alembic'] += ' -lAlembicAbcCoreAbstract'
		dll.linkmap['alembic'] += ' -lAlembicAbcCoreHDF5'
		dll.linkmap['alembic'] += ' -lAlembicAbcGeom'
#		dll.linkmap['alembic'] += ' -lAlembicAbcMaterial'

		deplibs = forge.corelibs+ [
					"fexSurfaceDLLib" ]

		forge.deps( [ dllname + "Lib" ], deplibs )

	forge.tests += [
		("inspect.exe",		"fexAlembicDL",				None,		None) ]

	module.Module('test')
